﻿using System;
using System.Collections;
using System.Collections.Generic;
using MoonTools.ECS.Collections;

namespace MoonTools.ECS;

public class Filter<T> where T : unmanaged
{
	private readonly Filter _underlying;

	internal Filter(Filter underlying)
	{
		_underlying = underlying;
	}

	public TypedEntityEnum<T> GetEnumerator()
	{
		return new TypedEntityEnum<T>(_underlying.Entities, _underlying.World);
	}
}

public ref struct TypedEntityEnum<T> : IEnumerator<(Entity entity, T)> where T : unmanaged
{
	private ReverseSpanEnumerator<Entity> _underlying;
	private readonly World _world;

	public TypedEntityEnum(ReverseSpanEnumerator<Entity> underlying, World world)
	{
		_underlying = underlying;
		_world = world;
	}

	public bool MoveNext()
	{
		if (!_underlying.MoveNext())
		{
			return false;
		}
		Entity currentEntity = _underlying.Current;
		Current = (currentEntity, _world.Get<T>(currentEntity));
		return true;
	}

	public void Reset()
	{
		throw new NotImplementedException();
	}

	public (Entity, T) Current { get; private set; }

	object IEnumerator.Current => Current;

	public void Dispose()
	{
	}
}

public class Filter<T1, T2> where T1 : unmanaged where T2 : unmanaged
{
	private readonly Filter _underlying;

	internal Filter(Filter underlying)
	{
		_underlying = underlying;
	}

	public TypedEntityEnum<T1, T2> GetEnumerator()
	{
		return new TypedEntityEnum<T1, T2>(_underlying.Entities, _underlying.World);
	}
}

public ref struct TypedEntityEnum<T1, T2> : IEnumerator<(Entity entity, T1, T2)> where T1 : unmanaged where T2 : unmanaged
{
	private ReverseSpanEnumerator<Entity> _underlying;
	private readonly World _world;

	public TypedEntityEnum(ReverseSpanEnumerator<Entity> underlying, World world)
	{
		_underlying = underlying;
		_world = world;
	}

	public bool MoveNext()
	{
		if (!_underlying.MoveNext())
		{
			return false;
		}
		Entity currentEntity = _underlying.Current;
		Current = (currentEntity, _world.Get<T1>(currentEntity), _world.Get<T2>(currentEntity));
		return true;
	}

	public void Reset()
	{
		throw new NotImplementedException();
	}

	public (Entity, T1, T2) Current { get; private set; }

	object IEnumerator.Current => Current;

	public void Dispose()
	{
	}
}

public class Filter<T1, T2, T3> where T1 : unmanaged where T2 : unmanaged where T3 : unmanaged
{
	private readonly Filter _underlying;

	internal Filter(Filter underlying)
	{
		_underlying = underlying;
	}

	public TypedEntityEnum<T1, T2, T3> GetEnumerator()
	{
		return new TypedEntityEnum<T1, T2, T3>(_underlying.Entities, _underlying.World);
	}
}

public ref struct TypedEntityEnum<T1, T2, T3> : IEnumerator<(Entity entity, T1, T2, T3)> where T1 : unmanaged where T2 : unmanaged where T3 : unmanaged
{
	private ReverseSpanEnumerator<Entity> _underlying;
	private readonly World _world;

	public TypedEntityEnum(ReverseSpanEnumerator<Entity> underlying, World world)
	{
		_underlying = underlying;
		_world = world;
	}

	public bool MoveNext()
	{
		if (!_underlying.MoveNext())
		{
			return false;
		}
		Entity currentEntity = _underlying.Current;
		Current = (currentEntity, _world.Get<T1>(currentEntity), _world.Get<T2>(currentEntity), _world.Get<T3>(currentEntity));
		return true;
	}

	public void Reset()
	{
		throw new NotImplementedException();
	}

	public (Entity, T1, T2, T3) Current { get; private set; }

	object IEnumerator.Current => Current;

	public void Dispose()
	{
	}
}

public class Filter
{
	internal World World;
	internal FilterSignature Signature;

	internal IndexableSet<Entity> EntitySet = new IndexableSet<Entity>();

	private bool IsDisposed;

	public ReverseSpanEnumerator<Entity> Entities => EntitySet.GetEnumerator();

	public bool Empty => EntitySet.Count == 0;
	public int Count => EntitySet.Count;

	/// WARNING: this WILL crash if the index is out of range!
	public Entity NthEntity(int index) => EntitySet[index];

	/// WARNING: this WILL crash if the filter is empty!
	public Entity RandomEntity => EntitySet[RandomManager.Next(EntitySet.Count)];
	public RandomEntityEnumerator EntitiesInRandomOrder => new RandomEntityEnumerator(this);

	internal Filter(World world, FilterSignature signature)
	{
		World = world;
		Signature = signature;
	}

	public void DestroyAllEntities()
	{
		foreach (var entity in EntitySet)
		{
			World.Destroy(entity);
		}
	}

	internal void Check(Entity entity)
	{
		foreach (var type in Signature.Included)
		{
			if (!World.Has(entity, type))
			{
				EntitySet.Remove(entity);
				return;
			}
		}

		foreach (var type in Signature.Excluded)
		{
			if (World.Has(entity, type))
			{
				EntitySet.Remove(entity);
				return;
			}
		}

		EntitySet.Add(entity);
	}

	internal void AddEntity(in Entity entity)
	{
		EntitySet.Add(entity);
	}

	internal void RemoveEntity(in Entity entity)
	{
		EntitySet.Remove(entity);
	}

	internal void Clear()
	{
		EntitySet.Clear();
	}

	public ref struct RandomEntityEnumerator
	{
		private Filter Filter;
		private LinearCongruentialEnumerator LinearCongruentialEnumerator;

		public RandomEntityEnumerator GetEnumerator() => this;

		internal RandomEntityEnumerator(Filter filter)
		{
			Filter = filter;
			LinearCongruentialEnumerator =
				RandomManager.LinearCongruentialSequence(filter.Count);
		}

		public bool MoveNext() => LinearCongruentialEnumerator.MoveNext();
		public Entity Current => Filter.NthEntity(LinearCongruentialEnumerator.Current);
	}

	protected virtual void Dispose(bool disposing)
	{
		if (!IsDisposed)
		{
			if (disposing)
			{
				EntitySet.Dispose();
			}

			IsDisposed = true;
		}
	}

	// // TODO: override finalizer only if 'Dispose(bool disposing)' has code to free unmanaged resources
	// ~Filter()
	// {
	//     // Do not change this code. Put cleanup code in 'Dispose(bool disposing)' method
	//     Dispose(disposing: false);
	// }

	public void Dispose()
	{
		// Do not change this code. Put cleanup code in 'Dispose(bool disposing)' method
		Dispose(disposing: true);
		GC.SuppressFinalize(this);
	}
}
